@@ -3,7 +3,7 @@
  *
  * The connection object - this file is part a of the C extension module.
  *
- * Copyright (c) 2025 by the PyGreSQL Development Team
+ * Copyright (c) 2026 by the PyGreSQL Development Team
  *
  * Please see the LICENSE.TXT file for specific restrictions.
  */
@@ -532,6 +532,12 @@ conn_describe_prepared(connObject *self, PyObject *args)
         query_obj->max_row = PQntuples(result);
         query_obj->num_fields = PQnfields(result);
         query_obj->col_types = get_col_types(result, query_obj->num_fields);
+        if (!query_obj->col_types) {
+            PQclear(result);
+            Py_DECREF(query_obj);
+            Py_DECREF(self);
+            return NULL;
+        }
         return (PyObject *)query_obj;
     }
     set_error(ProgrammingError, "Cannot describe prepared statement",
@@ -716,9 +722,9 @@ static PyObject *
 conn_inserttable(connObject *self, PyObject *args, PyObject *kwds)
 {
     PGresult *result;
-    char *table, *buffer, *bufpt, *bufmax, *s, *t;
+    char *table, *s, *t;
+    struct CharBuffer buffer = {0};
     int freeze = 0, encoding, ret;
-    size_t bufsiz;
     PyObject *rows, *iter_row, *item, *columns = NULL;
     Py_ssize_t i, j, m, n;
 
@@ -775,18 +781,16 @@ conn_inserttable(connObject *self, PyObject *args, PyObject *kwds)
         n = -1; /* number of columns not yet known */
     }
 
-    /* allocate buffer */
-    if (!(buffer = PyMem_Malloc(MAX_BUFFER_SIZE))) {
+    encoding = PQclientEncoding(self->cnx);
+
+    /* pre-allocate some memory for the query buffer */
+    if (!init_char_buffer(&buffer, 4096)) {
         Py_DECREF(iter_row);
         return PyErr_NoMemory();
     }
 
-    encoding = PQclientEncoding(self->cnx);
-
     /* starts query */
-    bufpt = buffer;
-    bufmax = bufpt + MAX_BUFFER_SIZE;
-    bufpt += snprintf(bufpt, (size_t)(bufmax - bufpt), "copy ");
+    ext_char_buffer_s(&buffer, "copy ");
 
     s = table;
     do {
@@ -794,18 +798,16 @@ conn_inserttable(connObject *self, PyObject *args, PyObject *kwds)
         if (!t)
             t = s + strlen(s);
         table = PQescapeIdentifier(self->cnx, s, (size_t)(t - s));
-        if (bufpt < bufmax)
-            bufpt += snprintf(bufpt, (size_t)(bufmax - bufpt), "%s", table);
+        ext_char_buffer_s(&buffer, table);
         PQfreemem(table);
         s = t;
-        if (*s && bufpt < bufmax)
-            *bufpt++ = *s++;
+        if (*s)
+            ext_char_buffer_c(&buffer, *s++);
     } while (*s);
 
     if (columns) {
         /* adds a string like f" ({','.join(columns)})" */
-        if (bufpt < bufmax)
-            bufpt += snprintf(bufpt, (size_t)(bufmax - bufpt), " (");
+        ext_char_buffer_s(&buffer, " (");
         for (j = 0; j < n; ++j) {
             PyObject *obj = PySequence_Fast_GET_ITEM(columns, j);
             Py_ssize_t slen;
@@ -817,7 +819,7 @@ conn_inserttable(connObject *self, PyObject *args, PyObject *kwds)
             else if (PyUnicode_Check(obj)) {
                 obj = get_encoded_string(obj, encoding);
                 if (!obj) {
-                    PyMem_Free(buffer);
+                    PyMem_Free(buffer.data);
                     Py_DECREF(iter_row);
                     return NULL; /* pass the UnicodeEncodeError */
                 }
@@ -826,37 +828,35 @@ conn_inserttable(connObject *self, PyObject *args, PyObject *kwds)
                 PyErr_SetString(
                     PyExc_TypeError,
                     "The third argument must contain only strings");
-                PyMem_Free(buffer);
+                PyMem_Free(buffer.data);
                 Py_DECREF(iter_row);
                 return NULL;
             }
             PyBytes_AsStringAndSize(obj, &col, &slen);
             col = PQescapeIdentifier(self->cnx, col, (size_t)slen);
             Py_DECREF(obj);
-            if (bufpt < bufmax)
-                bufpt += snprintf(bufpt, (size_t)(bufmax - bufpt), "%s%s", col,
-                                  j == n - 1 ? ")" : ",");
+            ext_char_buffer_s(&buffer, col);
             PQfreemem(col);
+            ext_char_buffer_c(&buffer, j == n - 1 ? ')' : ',');
         }
     }
-    if (bufpt < bufmax)
-        bufpt += snprintf(bufpt, (size_t)(bufmax - bufpt), " from stdin");
-    if (freeze && bufpt < bufmax) {
-        bufpt += snprintf(bufpt, (size_t)(bufmax - bufpt), " freeze");
-    }
+    ext_char_buffer_s(&buffer, " from stdin");
+    if (freeze)
+        ext_char_buffer_s(&buffer, " freeze");
+    ext_char_buffer_c(&buffer, '\0');
 
-    if (bufpt >= bufmax) {
-        PyMem_Free(buffer);
+    if (buffer.error) {
+        PyMem_Free(buffer.data);
         Py_DECREF(iter_row);
         return PyErr_NoMemory();
     }
 
     Py_BEGIN_ALLOW_THREADS
-    result = PQexec(self->cnx, buffer);
+    result = PQexec(self->cnx, buffer.data);
     Py_END_ALLOW_THREADS
 
     if (!result || PQresultStatus(result) != PGRES_COPY_IN) {
-        PyMem_Free(buffer);
+        PyMem_Free(buffer.data);
         Py_DECREF(iter_row);
         PyErr_SetString(PyExc_ValueError, PQerrorMessage(self->cnx));
         return NULL;
@@ -871,8 +871,7 @@ conn_inserttable(connObject *self, PyObject *args, PyObject *kwds)
 
         if (!(PyTuple_Check(columns) || PyList_Check(columns))) {
             PQputCopyEnd(self->cnx, "Invalid arguments");
-            PyMem_Free(buffer);
-            Py_DECREF(columns);
+            PyMem_Free(buffer.data);
             Py_DECREF(columns);
             Py_DECREF(iter_row);
             PyErr_SetString(
@@ -887,7 +886,7 @@ conn_inserttable(connObject *self, PyObject *args, PyObject *kwds)
         }
         else if (j != n) {
             PQputCopyEnd(self->cnx, "Invalid arguments");
-            PyMem_Free(buffer);
+            PyMem_Free(buffer.data);
             Py_DECREF(columns);
             Py_DECREF(iter_row);
             PyErr_SetString(
@@ -896,65 +895,54 @@ conn_inserttable(connObject *self, PyObject *args, PyObject *kwds)
             return NULL;
         }
 
-        /* builds insert line */
-        bufpt = buffer;
-        bufsiz = MAX_BUFFER_SIZE - 1;
+        /* empty buffer while keeping allocated memory */
+        buffer.size = 0;
+
+        /* build insert line */
 
         for (j = 0; j < n; ++j) {
-            if (j) {
-                *bufpt++ = '\t';
-                --bufsiz;
-            }
+            if (j)
+                ext_char_buffer_c(&buffer, '\t');
 
             item = PySequence_Fast_GET_ITEM(columns, j);
 
             /* convert item to string and append to buffer */
             if (item == Py_None) {
-                if (bufsiz > 2) {
-                    *bufpt++ = '\\';
-                    *bufpt++ = 'N';
-                    bufsiz -= 2;
-                }
-                else
-                    bufsiz = 0;
+                ext_char_buffer_c(&buffer, '\\');
+                ext_char_buffer_c(&buffer, 'N');
             }
             else if (PyBytes_Check(item)) {
                 const char *t = PyBytes_AsString(item);
 
-                while (*t && bufsiz) {
+                while (*t) {
                     switch (*t) {
                         case '\\':
-                            *bufpt++ = '\\';
-                            if (--bufsiz)
-                                *bufpt++ = '\\';
+                            ext_char_buffer_c(&buffer, '\\');
+                            ext_char_buffer_c(&buffer, '\\');
                             break;
                         case '\t':
-                            *bufpt++ = '\\';
-                            if (--bufsiz)
-                                *bufpt++ = 't';
+                            ext_char_buffer_c(&buffer, '\\');
+                            ext_char_buffer_c(&buffer, '\t');
                             break;
                         case '\r':
-                            *bufpt++ = '\\';
-                            if (--bufsiz)
-                                *bufpt++ = 'r';
+                            ext_char_buffer_c(&buffer, '\\');
+                            ext_char_buffer_c(&buffer, '\r');
                             break;
                         case '\n':
-                            *bufpt++ = '\\';
-                            if (--bufsiz)
-                                *bufpt++ = 'n';
+                            ext_char_buffer_c(&buffer, '\\');
+                            ext_char_buffer_c(&buffer, '\n');
                             break;
                         default:
-                            *bufpt++ = *t;
+                            ext_char_buffer_c(&buffer, *t);
                     }
                     ++t;
-                    --bufsiz;
                 }
             }
             else if (PyUnicode_Check(item)) {
                 PyObject *s = get_encoded_string(item, encoding);
                 if (!s) {
                     PQputCopyEnd(self->cnx, "Encoding error");
-                    PyMem_Free(buffer);
+                    PyMem_Free(buffer.data);
                     Py_DECREF(item);
                     Py_DECREF(columns);
                     Py_DECREF(iter_row);
@@ -963,33 +951,28 @@ conn_inserttable(connObject *self, PyObject *args, PyObject *kwds)
                 else {
                     const char *t = PyBytes_AsString(s);
 
-                    while (*t && bufsiz) {
+                    while (*t) {
                         switch (*t) {
                             case '\\':
-                                *bufpt++ = '\\';
-                                if (--bufsiz)
-                                    *bufpt++ = '\\';
+                                ext_char_buffer_c(&buffer, '\\');
+                                ext_char_buffer_c(&buffer, '\\');
                                 break;
                             case '\t':
-                                *bufpt++ = '\\';
-                                if (--bufsiz)
-                                    *bufpt++ = 't';
+                                ext_char_buffer_c(&buffer, '\\');
+                                ext_char_buffer_c(&buffer, '\t');
                                 break;
                             case '\r':
-                                *bufpt++ = '\\';
-                                if (--bufsiz)
-                                    *bufpt++ = 'r';
+                                ext_char_buffer_c(&buffer, '\\');
+                                ext_char_buffer_c(&buffer, '\r');
                                 break;
                             case '\n':
-                                *bufpt++ = '\\';
-                                if (--bufsiz)
-                                    *bufpt++ = 'n';
+                                ext_char_buffer_c(&buffer, '\\');
+                                ext_char_buffer_c(&buffer, '\n');
                                 break;
                             default:
-                                *bufpt++ = *t;
+                                ext_char_buffer_c(&buffer, *t);
                         }
                         ++t;
-                        --bufsiz;
                     }
                     Py_DECREF(s);
                 }
@@ -998,68 +981,59 @@ conn_inserttable(connObject *self, PyObject *args, PyObject *kwds)
                 PyObject *s = PyObject_Str(item);
                 const char *t = PyUnicode_AsUTF8(s);
 
-                while (*t && bufsiz) {
-                    *bufpt++ = *t++;
-                    --bufsiz;
-                }
+                ext_char_buffer_s(&buffer, t);
                 Py_DECREF(s);
             }
             else {
                 PyObject *s = PyObject_Repr(item);
                 const char *t = PyUnicode_AsUTF8(s);
 
-                while (*t && bufsiz) {
+                while (*t) {
                     switch (*t) {
                         case '\\':
-                            *bufpt++ = '\\';
-                            if (--bufsiz)
-                                *bufpt++ = '\\';
+                            ext_char_buffer_c(&buffer, '\\');
+                            ext_char_buffer_c(&buffer, '\\');
                             break;
                         case '\t':
-                            *bufpt++ = '\\';
-                            if (--bufsiz)
-                                *bufpt++ = 't';
+                            ext_char_buffer_c(&buffer, '\\');
+                            ext_char_buffer_c(&buffer, '\t');
                             break;
                         case '\r':
-                            *bufpt++ = '\\';
-                            if (--bufsiz)
-                                *bufpt++ = 'r';
+                            ext_char_buffer_c(&buffer, '\\');
+                            ext_char_buffer_c(&buffer, '\r');
                             break;
                         case '\n':
-                            *bufpt++ = '\\';
-                            if (--bufsiz)
-                                *bufpt++ = 'n';
+                            ext_char_buffer_c(&buffer, '\\');
+                            ext_char_buffer_c(&buffer, '\n');
                             break;
                         default:
-                            *bufpt++ = *t;
+                            ext_char_buffer_c(&buffer, *t);
                     }
                     ++t;
-                    --bufsiz;
                 }
                 Py_DECREF(s);
             }
-
-            if (bufsiz <= 0) {
-                PQputCopyEnd(self->cnx, "Memory error");
-                PyMem_Free(buffer);
-                Py_DECREF(columns);
-                Py_DECREF(iter_row);
-                return PyErr_NoMemory();
-            }
         }
 
         Py_DECREF(columns);
 
-        *bufpt++ = '\n';
+        /* terminate line */
+        ext_char_buffer_c(&buffer, '\n');
+        if (buffer.error) {
+            PQputCopyEnd(self->cnx, "Memory error");
+            PyMem_Free(buffer.data);
+            Py_DECREF(iter_row);
+            return PyErr_NoMemory();
+        }
 
-        /* sends data */
-        ret = PQputCopyData(self->cnx, buffer, (int)(bufpt - buffer));
+        /* send data */
+        ret = PQputCopyData(self->cnx, buffer.data, (int)buffer.size);
         if (ret != 1) {
             char *errormsg = ret == -1 ? PQerrorMessage(self->cnx)
                                        : "Data cannot be queued";
             PyErr_SetString(PyExc_IOError, errormsg);
             PQputCopyEnd(self->cnx, errormsg);
-            PyMem_Free(buffer);
+            PyMem_Free(buffer.data);
             Py_DECREF(iter_row);
             return NULL;
         }
@@ -1067,7 +1041,7 @@ conn_inserttable(connObject *self, PyObject *args, PyObject *kwds)
 
     Py_DECREF(iter_row);
     if (PyErr_Occurred()) {
-        PyMem_Free(buffer);
+        PyMem_Free(buffer.data);
         return NULL; /* pass the iteration error */
     }
 
@@ -1075,11 +1049,11 @@ conn_inserttable(connObject *self, PyObject *args, PyObject *kwds)
     if (ret != 1) {
         PyErr_SetString(PyExc_IOError, ret == -1 ? PQerrorMessage(self->cnx)
                                                  : "Data cannot be queued");
-        PyMem_Free(buffer);
+        PyMem_Free(buffer.data);
         return NULL;
     }
 
-    PyMem_Free(buffer);
+    PyMem_Free(buffer.data);
 
     Py_BEGIN_ALLOW_THREADS
     result = PQgetResult(self->cnx);
