@@ -3,7 +3,7 @@
  *
  * Internal functions - this file is part a of the C extension module.
  *
- * Copyright (c) 2025 by the PyGreSQL Development Team
+ * Copyright (c) 2026 by the PyGreSQL Development Team
  *
  * Please see the LICENSE.TXT file for specific restrictions.
  */
@@ -1493,3 +1493,72 @@ notice_receiver(void *arg, const PGresult *res)
     }
     PyGILState_Release(gstate);
 }
+
+/* Pre-allocate some memory for a char buffer and return success status. */
+static int
+init_char_buffer(struct CharBuffer *buf, size_t initial_size)
+{
+    buf->size = 0;
+    buf->data = PyMem_Malloc(initial_size);
+    if (buf->data) {
+        buf->max_size = initial_size;
+        buf->error = 0;
+    }
+    else {
+        buf->max_size = 0;
+        buf->error = 1;
+    }
+    return !buf->error;
+}
+
+/* Extend char buffer with given string.
+   Note: We do not assume or guarantee that the buffer is zero-terminated. */
+static void
+ext_char_buffer_s(struct CharBuffer *buf, const char *s)
+{
+    size_t len = strlen(s), need;
+
+    if (!len || buf->error)
+        return;
+
+    if ((need = buf->size + len) >= buf->max_size) {
+        void *tmp;
+
+        if (buf->max_size < 1024 * 1024) {
+            /* allocate powers of two unless it's large */
+            size_t double_size = 2 * buf->max_size;
+            if (double_size >= need) /* overflow check */
+                need = double_size;
+        }
+
+        tmp = PyMem_Realloc(buf->data, need);
+        if (!tmp) {
+            buf->error = 1;
+            return;
+        }
+
+        buf->data = tmp;
+        buf->max_size = need;
+    }
+
+    memcpy(buf->data + buf->size, s, len);
+    buf->size += len;
+}
+
+/* Extend char buffer with given character */
+static void
+ext_char_buffer_c(struct CharBuffer *buf, char c)
+{
+    if (buf->error)
+        return;
+
+    if (buf->size >= buf->max_size) { /* buffer is full? */
+        /* slow path dealing with reallocation */
+        char tmp[2] = {c ? c : '\n', '\0'}; /* allow adding a zero-byte */
+        ext_char_buffer_s(buf, tmp);
+        if (!c)
+            buf->data[buf->size - 1] = '\0'; /* fix zero-byte */
+    }
+    else
+        buf->data[buf->size++] = c;
+}
