@@ -3,7 +3,7 @@
  *
  * This is the main file for the C extension module.
  *
- * Copyright (c) 2025 by the PyGreSQL Development Team
+ * Copyright (c) 2026 by the PyGreSQL Development Team
  *
  * Please see the LICENSE.TXT file for specific restrictions.
  */
@@ -53,8 +53,7 @@ static const char *PyPgVersion = TOSTRING(PYGRESQL_VERSION);
 #define QUERY_MOVENEXT 3
 #define QUERY_MOVEPREV 4
 
-#define MAX_BUFFER_SIZE 65536 /* maximum transaction size */
-#define MAX_ARRAY_DEPTH 16    /* maximum allowed depth of an array */
+#define MAX_ARRAY_DEPTH 16 /* maximum allowed depth of an array */
 
 /* MODULE GLOBAL VARIABLES */
 
@@ -158,6 +157,18 @@ typedef struct {
 } largeObject;
 #define is_largeObject(v) (PyType(v) == &largeType)
 
+/*
+   A buffer for character data with routines to handle resizing.
+   This is inspired by libpq's PQExpBufferData.
+   The buffer can be extended with the extend_char_buffer_s/x() functions.
+*/
+struct CharBuffer {
+    char *data;      /* actual string data */
+    size_t size;     /* current size of data */
+    size_t max_size; /* allocated size */
+    int error;       /* error flag (invalid data) */
+};
+
 /* Internal functions */
 #include "pginternal.c"
 
